R.<u1,u2,v1,v2> = QQ[]

## Bases for sym^0 and sym^4
s0b1 = u1**2*v2**2 - 2*u1*u2*v1*v2 + u2**2*v1**2
s4b1 = u1**2*v1**2
s4b2 = 1/2 * (u1**2*v1*v2 + u1*u2*v1**2)
s4b3 = 1/6 * (u2**2*v1**2 + 4*u1*u2*v1*v2 + u1**2*v2**2)
s4b4 = 1/2 * (u2**2*v1*v2 + u1*u2*v2**2)
s4b5 = u2**2*v2**2

## quadratic forms of diagonal up to prec
prec = 7
qfs = [(n,r,m) for n in range(prec) for m in range(prec)
               for r in range(-isqrt(4*n*m),isqrt(4*n*m)+1)]


## we use a combinatorial algebra to multiply Siegel modular forms
class FEAlgebra(CombinatorialFreeModule):
  def __init__(self):
     super(FEAlgebra,self).__init__(R,qfs,category=AlgebrasWithBasis(R))

  def product_on_basis(self,i,j):
    ij = tuple(map(operator.add,i,j))
    if ij[0] >= prec or ij[2] >= prec:
      return self.zero()
    else:
      return self(ij)

  def one(self):
    return self((0,0,0))

FE = FEAlgebra()

def reduce_qf(qf):
  bqf = BinaryQF(qf)

  if bqf.discriminant() == 0:
    n = gcd([bqf[0],bqf[1],bqf[2]])
    return (n,0,0)
  else:
    bqf = bqf.reduced_form()
    return (bqf[2], abs(bqf[1]), bqf[0])

def from_dict(d):
  res = FE.zero()

  for qf in qfs:
    res += d[reduce_qf(qf)] * FE(qf)

  return res


## differentials with respect to tau, z, and tau'
df1 = lambda f: sum(qf[0]*f[qf]*FE(qf) for qf in qfs)
df2 = lambda f: sum(1/2*qf[1]*f[qf]*FE(qf) for qf in qfs)
df3 = lambda f: sum(qf[2]*f[qf]*FE(qf) for qf in qfs)

def ibukiyama_bracket(f, g, k, l):
  ## Ibukiyama uses a different differential in his explicit expression
  df2 = lambda f: sum(qf[1]*f[qf]*FE(qf) for qf in qfs)

  t11 = 1/2*l*(l+1) * g
  t12 = ( s4b1 * df1(df1(f))
        + s4b2 * 2*df1(df2(f))
        + s4b3 * ( df2(df2(f)) + 2*df1(df3(f)) )
        + s4b4 * 2*df2(df3(f))
        + s4b5 * df3(df3(f)) )
  t21 = -(k+1)*(l+1)
  t22 = ( s4b1 * df1(f)*df1(g)
        + s4b2 * ( df1(f)*df2(g) + df2(f)*df1(g) )
        + s4b3 * ( df1(f)*df3(g) + df2(f)*df2(g) + df3(f)*df1(g) )
        + s4b4 * ( df3(f)*df2(g) + df2(f)*df3(g) )
        + s4b5 * df3(f)*df3(g) )
  t31 = 1/2*k*(k+1) * f
  t32 = ( s4b1 * df1(df1(g))
        + s4b2 * 2*df1(df2(g))
        + s4b3 * ( df2(df2(g)) + 2*df1(df3(g)) )
        + s4b4 * 2*df2(df3(g))
        + s4b5 * df3(df3(g)) )
 
  return t11*t12 + t21*t22 + t31*t32 

## three helper functions used to define f_RS
def dtaudtau(f):
  return ( u1**2 * df1(f) + (2*u1*u2) * df2(f) + u2**2 * df3(f) ) \
       * ( v1**2 * df1(f) + (2*v1*v2) * df2(f) + v2**2 * df3(f) )

def ddtau(f):
  return ( u1**2      * v1**2      * df1(df1(f))
         + (2*u1*u2)  * v1**2      * df2(df1(f))
         + u2**2      * v1**2      * df3(df1(f))
         + u1**2      * (2*v1*v2)  * df1(df2(f))
         + (2*u1*u2)  * (2*v1*v2)  * df2(df2(f))
         + u2**2      * (2*v1*v2)  * df3(df2(f))
         + u1**2      * v2**2      * df1(df3(f))
         + (2*u1*u2)  * v2**2      * df2(df3(f))
         + u2**2      * v2**2      * df3(df3(f)) )
 
def tdtaudtau(f):
  return ( u1**2 * v1**2
           * df1(f) * df1(f)
         + 1/2 * ( (2*u1*u2) * v1**2 + u1**2 * (2*v1*v2) )
           * df2(f) * df1(f)
         + 1/4 * (2*u1*u2) * (2*v1*v2)
           * df3(f) * df1(f)
         + 1/2 * ( u1**2 * (2*v1*v2) + (2*u1*u2) * v1**2 )
           * df1(f) * df2(f)
         + ( 1/2 * (2*u1*u2) * (2*v1*v2) + u1**2*v2**2 + u2**2*v1**2 )
           * df2(f) * df2(f)
         + 1/2 * ( u2**2 * (2*v1*v2) + (2*u1*u2) * v2**2 )
           * df3(f) * df2(f)
         + 1/4 * (2*u1*u2) * (2*v1*v2)
           * df1(f) * df3(f)
         + 1/2 * ( (2*u1*u2) * v2**2 + u2**2 * (2*v1*v2) )
           * df2(f) * df3(f)
         + u2**2 * v2**2
           * df3(f) * df3(f) )



## expansions computed in fourier_expansion.sage


e4 = from_dict(dict(load("e4_fe.sobj")))
e6 = from_dict(dict(load("e6_fe.sobj")))
phi2_10 = from_dict(dict(load("phi2_10_fe.sobj")))
e12 = from_dict(dict(load("e12_fe.sobj")))
ed = from_dict(dict(load("ed_fe.sobj")))
phi2_12 = from_dict(dict(load("phi2_12_fe.sobj")))

# Ibukiyama's form of weight det10 sym4
ib4 = ibukiyama_bracket(e4,e6,4,6)

# product of f_RS and phi2_10^2
fRS_phi2_10_sq = ( -1/10 * dtaudtau(phi2_10)
                 + 1/10 * phi2_10 * ddtau(phi2_10)
                 - 1/100 * tdtaudtau(phi2_10) )

# a basis for the space of Siegel modular forms of
# weigth det10 sym2 sym2
mbasis = [ ib4,
           s0b1*e12, s0b1*ed, 
           s0b1*phi2_12 ]
mbasis_phi2_10 = [phi2_10 * b for b in mbasis]

## coefficients of u1**2*v1**2 and u2**2*v1**2
## at (2,-2,2), (2,-1,2), and (2,0,2)
to_vec = lambda f: vector([ f[(2,-2,2)].coefficient(u1**2*v1**2),
                            f[(2,-2,2)].coefficient(u2**2*v1**2),
                            f[(2,-1,2)].coefficient(u1**2*v1**2),
                            f[(2,-1,2)].coefficient(u2**2*v1**2),
                            f[(2, 0,2)].coefficient(u1**2*v1**2),
                            f[(2, 0,2)].coefficient(u2**2*v1**2) ])

## coefficients of phi2_10^2 f_RS with respect to the above
## basis of modular forms
mcoeffs = matrix(QQ, map(to_vec, mbasis_phi2_10)) \
          .solve_left( to_vec(fRS_phi2_10_sq) )
assert 0 == fRS_phi2_10_sq - sum(map(operator.mul,mcoeffs,mbasis_phi2_10))

## Fourier coefficients of f_RS
fRS_phi2_10 = sum(map(operator.mul,mcoeffs,mbasis))
save(dict(list(fRS_phi2_10)), "fRS_phi2_10_fe.sobj")

